/*
 * (C) Copyright 2000
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

/*
 * Boot support
 */
#include <ppcboot.h>
#include <command.h>
#include <cmd_boot.h>
#include <s_record.h>


static ulong load_serial (ulong offset);
static int read_record (char *buf, ulong len);

void do_bdinfo (cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
	int i;
	ulong ip = bd->bi_ip_addr;

	printf ("  memstart    = 0x%08lx\n",  bd->bi_memstart   );
	printf ("  memsize     = 0x%08lx\n",  bd->bi_memsize    );
	printf ("  flashstart  = 0x%08lx\n",  bd->bi_flashstart );
	printf ("  flashsize   = 0x%08lx\n",  bd->bi_flashsize  );
	printf ("  flashoffset = 0x%08lx\n",  bd->bi_flashoffset);
	printf ("  sramstart   = 0x%08lx\n",  bd->bi_sramstart  );
	printf ("  sramsize    = 0x%08lx\n",  bd->bi_sramsize   );
#ifdef	CONFIG_8xx
	printf ("  immr_base   = 0x%08lx\n",  bd->bi_immr_base  );
#endif
	printf ("  bootflags   = 0x%08lx\n",  bd->bi_bootflags  );
	printf ("  intfreq     = %6ld MHz\n", bd->bi_intfreq    );
	printf ("  busfreq     = %6ld MHz\n", bd->bi_busfreq    );
	printf ("  ethaddr     =");
	for (i=0; i<6; ++i) {
		printf ("%c%02X", i ? ':' : ' ', bd->bi_enetaddr[i]);
	}
	printf ("\n  IP addr     =");
	for (i=0; i<4; ++i) {
		printf ("%c%ld", i ? '.' : ' ', (ip >> 24) & 0xFF);
		ip <<= 8;
	}
	printf ("\n  baudrate    = %6ld bps\n", bd->bi_baudrate   );
	printf ("  getc        = 0x%08lx\n",(ulong)bd->bi_serial_io.getc);
	printf ("  tstc        = 0x%08lx\n",(ulong)bd->bi_serial_io.tstc);
	printf ("  putc        = 0x%08lx\n",(ulong)bd->bi_serial_io.putc);
	printf ("  printf      = 0x%08lx\n",(ulong)bd->bi_serial_io.printf);
	printf ("  install_hdlr= 0x%08lx\n",(ulong)bd->bi_interrupt.install_hdlr);
	printf ("  free_hdr    = 0x%08lx\n",(ulong)bd->bi_interrupt.free_hdlr);
}

void do_go (cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
	ulong	addr, rc;

	if (argc < 2) {
		printf ("Usage:\n%s\n", cmdtp->usage);
		return;
	}

	asc_to_hex(argv[1], &addr);

	printf ("## Starting application at 0x%08lx ...\n", addr);

	/*
	 * pass address parameter as argv[0] (aka command name),
	 * and all remaining args
	 */
	rc = ((ulong (*)(bd_t *, int, char *[]))addr) (bd, --argc, &argv[1]);

	printf ("## Application terminated, rc = 0x%lx\n", rc);
}

void do_load_serial (cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
	ulong offset = 0;
	ulong addr;
	int i;

	if (argc == 2) {
		asc_to_hex(argv[1], &offset);
	}

	command_repeat_off();

	printf ("## Ready for S-Record download ...\n");

	addr = load_serial (offset);

	/*
	 * Gather any trailing characters (for instance, the ^D which
	 * is sent by 'cu' after sending a file), and give the
	 * box some time (100 * 1 ms)
	 */
	for (i=0; i<100; ++i) {
		if (serial_tstc()) {
			(void) serial_getc();
		}
		udelay(1000);
	}

	if (addr == ~0) {
		printf ("## S-Record download aborted\n");
	} else {
		printf ("## Start Addr = 0x%08lx\n", addr);
	}
}

static ulong
load_serial (ulong offset)
{
	char	record[SREC_MAXRECLEN + 1];	/* buffer for one S-Record	*/
	char	binbuf[SREC_MAXBINLEN];		/* buffer for binary data	*/
	int	binlen;				/* no. of data bytes in S-Rec.	*/
	int	type;				/* return code for record type	*/
	ulong	addr;				/* load address from S-Record	*/

	while (read_record(record, SREC_MAXRECLEN + 1) >= 0) {
		type = srec_decode (record, &binlen, &addr, binbuf);

		if (type < 0) {
			return (~0);		/* Invalid S-Record		*/
		}

		switch (type) {
		case SREC_DATA2:
		case SREC_DATA3:
		case SREC_DATA4:
		    if (addr2info(addr+offset)) {
			int rc; 

			switch (rc=flash_write((uchar *)binbuf,addr+offset,binlen))
			{
			case 0:	break;
			case 1: printf ("\nError: Timeout writing to Flash\n");
				return (~0);
			case 2:	printf ("\nError: Flash not Erased\n");
				return (~0);
			case 4: printf ("\nError: Can't write to protected Flash sectors\n");
				return (~0);
			default:
				printf ("%s[%d] FIXME: rc=%d\n",
					__FILE__,__LINE__,rc);
				return (~0);
			}
		    } else {
			memcpy ((char *)(addr+offset), binbuf, binlen);
		    }
		    break;
		case SREC_END2:
		case SREC_END3:
		case SREC_END4:
		    return (addr);
		case SREC_START:
		    break;
		default:
		    break;
		}
	}

	return (~0);			/* Download aborted		*/
}

static int
read_record (char *buf, ulong len)
{
	char *p;
	char c;

	--len;	/* always leave room for terminating '\0' byte */

	for (p=buf; p < buf+len; ++p) {
		c = serial_getc();		/* read character		*/
		serial_putc (c);		/* ... and echo it		*/

		switch (c) {
		case '\r':
		case '\n':
			*p = '\0';
			return (p - buf);
		case '\0':
		case 0x03:			/* ^C - Control C		*/
			return (-1);
		default:
			*p = c;
		}
	}

	/* line too long - truncate */
	*p = '\0';
	return (p - buf);
}
