/*
 * (C) Copyright 2000
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#ifndef _PPCBOOT_H_
#define _PPCBOOT_H_	1

#define CONFIG_8xx 1		/* needed for Linux kernel header files */

#undef	_LINUX_CONFIG_H
#define _LINUX_CONFIG_H 1	/* avoid reading Linux autoconf.h file	*/

#include "config.h"
#include <linux/bitops.h>
#include <linux/types.h>
#include <linux/string.h>
#ifdef	CONFIG_8xx
#include <asm/8xx_immap.h>
#endif

typedef unsigned char		uchar;
typedef volatile unsigned long	vu_long;

#include <flash.h>

#ifdef	DEBUG
#define debug(fmt,args...)	printf (fmt ,##args)
#else
#define debug(fmt,args...)
#endif

typedef	void (interrupt_handler_t)(void *);

typedef struct serial_io {
	int	(*getc)(void);
	int	(*tstc)(void);
	void	(*putc)(const char c);
	void	(*printf)(const char *fmt, ...);
} serial_io_t;

typedef struct intr_util {
	void	(*install_hdlr)(int, interrupt_handler_t *, void *);
	void	(*free_hdlr)(int);
} intr_util_t;


/* A Board Information structure that is given to a program when
 * ppcboot starts it up.
 */
typedef struct bd_info {
	 unsigned long	bi_memstart;	/* start of  DRAM memory		*/
	 unsigned long	bi_memsize;	/* size	 of  DRAM memory in bytes	*/
	 unsigned long	bi_flashstart;	/* start of FLASH memory		*/
	 unsigned long	bi_flashsize;	/* size	 of FLASH memory		*/
	 unsigned long	bi_flashoffset; /* reserved area for startup monitor	*/
	 unsigned long	bi_sramstart;	/* start of  SRAM memory		*/
	 unsigned long	bi_sramsize;	/* size	 of  SRAM memory		*/
#ifdef	CONFIG_8xx
	 unsigned long	bi_immr_base;	/* base of IMMR register		*/
#endif
	 unsigned long	bi_bootflags;	/* boot / reboot flag (for LynxOS)	*/
	 unsigned long	bi_ip_addr;	/* IP Address				*/
	 unsigned char	bi_enetaddr[6]; /* Ethernet adress			*/
	 unsigned char	bi_reserved[2]; /* -- just for alignment --		*/
	 unsigned long	bi_intfreq;	/* Internal Freq, in MHz		*/
	 unsigned long	bi_busfreq;	/* Bus Freq, in MHz			*/
	 unsigned long	bi_baudrate;	/* Console Baudrate			*/
	 serial_io_t	bi_serial_io;	/* Addr of monitor fnc for Console I/O	*/
	 intr_util_t	bi_interrupt;	/* Addr of monitor fnc for Interrupts	*/
} bd_t;

/*
 * Function Prototypes
 */

void	main_loop     (bd_t *);
void	hang	      (void);
void	udelay	      (unsigned long);


/* */
long int initdram (int);
void	display_options	 (void);

/* common/main.c */
int readline	      (const char *const prompt);
void command_repeat_off (void);

/* common/board.c */
void	board_init_f  (ulong);
void	board_init_r  (bd_t *, ulong);
int	set_board_info(bd_t *, uint);
int	checkboard    (void);
int	checkflash    (void);
int	checkdram     (void);

/* common/cmd_nvedit.c */
char *getenv (uchar *);

/* board/flash.c */
ulong flash_get_size (vu_long *addr, flash_info_t *info);
int flash_write (uchar *, ulong, ulong);
flash_info_t *addr2info (ulong);

/* $(CPU)/serial.c */
void	serial_init   (ulong);
void	serial_putc   (const char);
void	serial_putstr (const char *);
void	serial_addr   (unsigned int);
int	serial_getc   (void);
int	serial_tstc   (void);

/* $(CPU)/start.S */
#ifdef	CONFIG_8xx
uint	get_immr      (uint);
#endif
uint	get_pvr	      (void);
uint	rd_ic_cst     (void);
void	wr_ic_cst     (uint);
void	wr_ic_adr     (uint);
uint	rd_dc_cst     (void);
void	wr_dc_cst     (uint);
void	wr_dc_adr     (uint);
int	icache_status (void);
void	icache_enable (void);
void	icache_disable(void);
int	dcache_status (void);
void	dcache_enable (void);
void	dcache_disable(void);
void	relocate_code (ulong, bd_t *, ulong);
ulong	get_endaddr   (void);
void	trap_init     (ulong);

/* $(CPU)/cpu.c */
int	checkcpu      (long);
int	checkicache   (void);
int	checkdcache   (void);
void	upmconfig     (unsigned int, unsigned int *, unsigned int);

/* $(CPU)/speed.c */
ulong	get_gclk_freq (void);
ulong	get_bus_freq  (ulong);

/* $(CPU)/cpu_init.c */
#ifdef	CONFIG_8xx
void	cpu_init_f    (volatile immap_t *immr);
#endif
void	cpu_init_r    (bd_t *bd);

/* $(CPU)/interrupts.c */
void	interrupt_init     (bd_t *bd);
void	timer_interrupt    (struct pt_regs *);
void	external_interrupt (struct pt_regs *);
void	cpm_install_handler(int, interrupt_handler_t *, void *);
void	cpm_free_handler   (int);
void	reset_timer	   (void);
ulong	get_timer	   (ulong base);
void	set_timer	   (ulong t);
void	enable_interrupts  (void);
int	disable_interrupts (void);

/* $(CPU)/traps.c */

/* ppc/vsprintf.c */
ulong	simple_strtoul(const char *cp,char **endp,unsigned int base);
long	simple_strtol(const char *cp,char **endp,unsigned int base);
uint	asc_to_hex(char *cp, ulong *val);
void	printf(const char *fmt, ...);
void	panic(const char *fmt, ...);
int	sprintf(char * buf, const char *fmt, ...);

/* ppc/crc32.c */
ulong crc32 (ulong, const unsigned char *, uint);

#endif	/* _PPCBOOT_H_ */
